@define-color bg_color rgba(0,0,0,0.2);
@define-color borders_color rgba(0,0,0,0.5);
@define-color shadow_color rgba(0,0,0,0.3);
@define-color accent_color #FF2E88;
@define-color banner_bg_color rgba(42, 123, 222, 0.3);

.keycap {
  min-width: 15px;
  min-height: 25px;
  margin-top: 2px;
  padding-bottom: 3px;
  padding-left: 6px;
  padding-right: 6px;
  background-color: @bg_color;
  border: 1px solid;
  border-radius: 5px;
  border-color: @borders_color;
  box-shadow: inset 0 -3px @shadow_color;
}

.bindings-section header label {
  color: @accent_color;
  margin-top: 26px;
  filter: unset;
}

.popover_menu_box > list > row {
  border-radius: 10px;
}

.popover_menu_box .setting_label {
  font-size: 85%;
  opacity: 0.8;
}

.popover_menu_box .menu_entry {
  font-weight: normal;
}

.transparent,
.bindings-section,
.bindings-section .view {
  background-color: transparent;
}

.banner {
  background-color: @banner_bg_color;
  font-weight: bold;
}
